/**
 * Legato is a configurable, lightweight web mapping client that can be
 * easily embedded into web pages and portals, CMS and individual web
 * applications. Legato is implemented in JavaScript and based on the
 * popular open source library OpenLayers.
 *
 * Copyright (C) 2010  disy Informationssysteme GmbH, http://www.disy.net
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.disy.ogc.wps.v_1_0_0.annotation;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import net.disy.ogc.ows.v_1_1_0.NoApplicableCodeException;
import net.disy.ogc.ows.v_1_1_0.OwsException;
import net.disy.ogc.wps.v_1_0_0.procedure.InputParameterDescription;
import net.disy.ogc.wps.v_1_0_0.procedure.OutputParameterDescription;
import net.disy.ogc.wps.v_1_0_0.procedure.WpsProcedure;
import net.disy.ogc.wps.v_1_0_0.util.WpsProcessUtilities;
import net.opengis.ows.v_1_1_0.MetadataType;

import org.apache.commons.collections15.CollectionUtils;
import org.apache.commons.collections15.Predicate;

@SuppressWarnings("nls")
public class AnnotationWpsProcedure implements WpsProcedure {
  private final Object annotatedObject;

  public AnnotationWpsProcedure(Object annotatedObject) {
    this.annotatedObject = annotatedObject;

    {
      Process process = annotatedObject.getClass().getAnnotation(Process.class);
      if (process == null) {
        throw new IllegalArgumentException(MessageFormat.format(
            "Class must be annotated with {0} annotation.",
            Process.class.getName()));
      }
      this.process = process;

    }

    {
      this.processMethod = WpsProcessUtilities.getSingleRequired(annotatedObject
          .getClass()
          .getMethods(), new Predicate<Method>() {
        @Override
        public boolean evaluate(Method object) {
          return object.isAnnotationPresent(ProcessMethod.class);
        }
      }, new IllegalArgumentException("Exactly one method must be annotated with "
          + ProcessMethod.class.getName()));
    }

    {
      List<InputParameterDescription> inputParameters = new ArrayList<InputParameterDescription>();
      Class<?>[] parameterTypes = processMethod.getParameterTypes();
      Annotation[][] parameterAnnotations = processMethod.getParameterAnnotations();
      for (int i = 0; i < parameterTypes.length; i++) {
        InputParameter inputParameter = getInputParameter(parameterAnnotations[i]);
        Annotation defaultValueAnnotation = getDefaultValue(parameterAnnotations[i]);

        InputParameterDescription inputParameterDescription = null;

        if (defaultValueAnnotation != null) {
          try {
            final Class<? extends DefaultValueFactory<?, ?>> defaultValueFactoryClass = defaultValueAnnotation
                .annotationType()
                .getAnnotation(DefaultValue.class)
                .createdBy();

            @SuppressWarnings("unchecked")
            DefaultValueFactory<Annotation, Object> defaultValueFactory = (DefaultValueFactory<Annotation, Object>) defaultValueFactoryClass
                .newInstance();

            Object defaultValue = defaultValueFactory.createDefaultValue(defaultValueAnnotation);
            inputParameterDescription = new AnnotationInputParameterDescription(
                parameterTypes[i],
                inputParameter,
                defaultValue);

          }
          catch (Exception ex) {
            throw new IllegalArgumentException("Could not create the default value instance.", ex);

          }
        }
        else {
          inputParameterDescription = new AnnotationInputParameterDescription(
              parameterTypes[i],
              inputParameter);

        }

        inputParameters.add(inputParameterDescription);
      }
      this.inputParameters = Collections.unmodifiableList(inputParameters);
    }

    {
      List<OutputParameterDescription> outputParameters = new ArrayList<OutputParameterDescription>();
      OutputParameter outputParameter = processMethod.getAnnotation(OutputParameter.class);
      if (outputParameter != null) {
        outputParameters.add(new AnnotationOutputParameterDescription(
            outputParameter,
            processMethod.getReturnType()));
      }

      this.outputParameters = Collections.unmodifiableList(outputParameters);
    }
  }

  private InputParameter getInputParameter(Annotation[] parameterAnnotations) {
    InputParameter inputParameter = (InputParameter) WpsProcessUtilities.getSingleRequired(
        parameterAnnotations,
        new Predicate<Annotation>() {
          @Override
          public boolean evaluate(Annotation value) {
            return value instanceof InputParameter;
          }
        },
        new IllegalStateException("Each parameter must be annotated with "
            + InputParameter.class.getName()));
    return inputParameter;
  }

  private Annotation getDefaultValue(Annotation[] parameterAnnotations) {
    return CollectionUtils.find(Arrays.asList(parameterAnnotations), new Predicate<Annotation>() {
      @Override
      public boolean evaluate(Annotation value) {
        return value.annotationType().isAnnotationPresent(DefaultValue.class);
      }
    });
  }

  @Override
  public String getId() {
    return getProcess().id();
  }

  @Override
  public String getTitle() {
    return getProcess().title();
  }

  @Override
  public String getAbstract() {
    return getProcess().description();
  }

  @Override
  public List<MetadataType> getMetadata() {
    return WpsProcessUtilities.getMetadata(getProcess().metadata());
  }

  private final Process process;

  private Process getProcess() {
    return process;
  }

  private final Method processMethod;

  private final List<InputParameterDescription> inputParameters;

  @Override
  public List<InputParameterDescription> getInputParameters() {
    return inputParameters;
  }

  private final List<OutputParameterDescription> outputParameters;

  @Override
  public List<OutputParameterDescription> getOutputParameters() {
    return outputParameters;
  }

  @Override
  public Object execute(Object[] arguments) throws OwsException {
    try {
      return processMethod.invoke(annotatedObject, arguments);
    }
    catch (InvocationTargetException e) {
      if (e.getCause() instanceof OwsException) {
        throw (OwsException) e.getCause();
      }

      throw new NoApplicableCodeException("Service threw exception", e.getCause());
    }
    catch (IllegalAccessException e) {
      throw new NoApplicableCodeException("Service not accessible", e);
    }
  }

}
