/**
 * Legato is a configurable, lightweight web mapping client that can be
 * easily embedded into web pages and portals, CMS and individual web
 * applications. Legato is implemented in JavaScript and based on the
 * popular open source library OpenLayers.
 *
 * Copyright (C) 2010  disy Informationssysteme GmbH, http://www.disy.net
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.disy.ogc.wps.v_1_0_0.converter;

import static net.disy.ogc.wps.v_1_0_0.util.Validate.*;
import static org.apache.commons.lang.Validate.*;

import java.util.HashMap;
import java.util.Map;

public class DefaultTypeConverterForSource<S> implements TypeConverterForSource<S> {
  private final Map<Class<?>, TypeConverter<S, ?>> typeConverters = new HashMap<Class<?>, TypeConverter<S, ?>>();
  private final Class<S> sourceClass;

  public DefaultTypeConverterForSource(Class<S> sourceClass) {
    notNull(sourceClass);
    this.sourceClass = sourceClass;
  }

  public void addConverter(TypeConverter<S, ?> typeConverter) {
    notNull(typeConverter);
    notContains(typeConverters, typeConverter.getDestinationClass());
    typeConverters.put(typeConverter.getDestinationClass(), typeConverter);
  }

  @Override
  public Class<S> getSourceClass() {
    return sourceClass;
  }

  @Override
  public <D> D convertTo(Class<D> destinationClass, S object) {
    notNull(destinationClass);
    TypeConverter<S, D> typeConverter = getRequiredTypeConverter(destinationClass);
    return typeConverter.convertTo(object);
  }

  @Override
  public <D> S convertFrom(Class<D> destinationClass, D object) {
    notNull(destinationClass);
    TypeConverter<S, D> typeConverter = getRequiredTypeConverter(destinationClass);
    return typeConverter.convertFrom(object);
  }

  @SuppressWarnings("unchecked")
  private <D> TypeConverter<S, D> getRequiredTypeConverter(Class<D> destinationClass) {
    if (!isConversionSupported(destinationClass)) {
      throw new IllegalArgumentException("Class '" //$NON-NLS-1$
          + destinationClass.getName()
          + "' is not supported."); //$NON-NLS-1$
    }

    return getTypeConverter(destinationClass);
  }

  private <D> TypeConverter<S, D> getTypeConverter(Class<D> destinationClass) {
    if (getSourceClass() == destinationClass) {
      return (TypeConverter<S, D>) new IdenticalConverter<S>(getSourceClass());
    }

    if (typeConverters.containsKey(destinationClass)) {
      return (TypeConverter<S, D>) typeConverters.get(destinationClass);
    }

    return null;
  }

  @Override
  public boolean isConversionSupported(Class<?> destinationClass) {
    notNull(destinationClass);
    return getTypeConverter(destinationClass) != null;
  }

}
