/**
 * Legato is a configurable, lightweight web mapping client that can be
 * easily embedded into web pages and portals, CMS and individual web
 * applications. Legato is implemented in JavaScript and based on the
 * popular open source library OpenLayers.
 *
 * Copyright (C) 2010  disy Informationssysteme GmbH, http://www.disy.net
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.disy.ogc.wps.v_1_0_0.procedure;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.disy.ogc.wps.v_1_0_0.model.DataType;
import net.disy.ogc.wps.v_1_0_0.model.DataTypeTypeVisitor;
import net.disy.ogc.wps.v_1_0_0.model.Format;
import net.disy.ogc.wps.v_1_0_0.model.LiteralType;
import net.opengis.ows.v_1_1_0.CodeType;
import net.opengis.wps.v_1_0_0.InputType;

@SuppressWarnings("nls")
public class InputArgumentBuilder {
  private final Map<String, List<InputType>> inputMap = new HashMap<String, List<InputType>>();
  private List<InputParameterDescription> inputParameters;
  private Set<String> parameterIds;
  private final WpsProcessContext context;

  public InputArgumentBuilder(WpsProcessContext context) {
    this.context = context;
  }

  public void setInputParameters(List<InputParameterDescription> inputParameters) {
    this.inputParameters = inputParameters;
    parameterIds = new HashSet<String>();
    for (InputParameterDescription inputParameter : inputParameters) {
      parameterIds.add(inputParameter.getId());
    }
  }

  public void addInput(InputType input) {
    final CodeType identifier = input.getIdentifier();
    if (identifier.getCodeSpace() != null) {
      throw new IllegalArgumentException("CodeSpace must be null.");
    }

    String parameterId = identifier.getValue();
    if (!parameterIds.contains(parameterId)) {
      throw new IllegalArgumentException("Unexpected InputParameter with identifier '"
          + parameterId
          + "'");
    }

    getInputs(parameterId).add(input);
  }

  private List<InputType> getInputs(String parameterId) {
    if (inputMap.containsKey(parameterId)) {
      return inputMap.get(parameterId);
    }

    List<InputType> inputList = new ArrayList<InputType>();
    inputMap.put(parameterId, inputList);
    return inputList;
  }

  public Object[] getArguments() {
    Object[] arguments = new Object[inputParameters.size()];
    for (int i = 0; i < inputParameters.size(); i++) {
      InputParameterDescription inputParameter = inputParameters.get(i);
      List<InputType> input = getInputs(inputParameter.getId());
      checkInputCount(inputParameter, input.size());
      arguments[i] = getInputArgument(input, inputParameter);
    }
    return arguments;
  }

  private void checkInputCount(InputParameterDescription inputParameter, int size) {
    if (!inputParameter.isOptional() && size == 0) {
      throw new IllegalArgumentException("inputParameter '"
          + inputParameter.getId()
          + "' is required.");
    }
  }

  private <S, D> Object getInputArgument(
      List<InputType> inputList,
      InputParameterDescription inputParameter) {
    Class<?> typeClass = inputParameter.getParameterClass();
    if (typeClass.isArray()) {
      Object array = Array.newInstance(typeClass.getComponentType(), inputList.size());
      for (int i = 0; i < inputList.size(); i++) {
        Object value = getInputValue(typeClass.getComponentType(), inputParameter, inputList.get(i));
        Array.set(array, i, value);
      }
      return array;
    }
    if (inputList.isEmpty()) {
      return null;
    }
    if (inputList.size() != 1) {
      throw new IllegalArgumentException("Input parameter count mismatch.");
    }

    return getInputValue(typeClass, inputParameter, inputList.get(0));
  }

  private <S, D> Object getInputValue(
      Class<D> theClass,
      InputParameterDescription inputParameter,
      InputType inputData) {
    @SuppressWarnings("unchecked")
    final DataType<S> dataType = (DataType<S>) context.getDataTypeHelper().getDataType(
        inputParameter,
        theClass);
    //    inputParameter.getParameterType().getDataType(
    //        context.getDataTypeRegistry(),
    //        theClass);

    final net.opengis.wps.v_1_0_0.DataType input = inputData.getData();
    class InputValueProvider implements DataTypeTypeVisitor {
      S unconvertedValue;

      @SuppressWarnings("unchecked")
      @Override
      public void visitBBox() {
        unconvertedValue = (S) input.getBoundingBoxData();
      }

      @Override
      public void visitComplex() {
        unconvertedValue = getComplexInputValue(dataType, input);
      }

      @Override
      public void visitLiteral() {
        unconvertedValue = getLiteralInputValue(dataType, input);
      }
    }

    InputValueProvider inputValueProvider = new InputValueProvider();
    dataType.getType().accept(inputValueProvider);
    return context.getConverterRegistry().convertTo(
        dataType.getTargetClass(),
        theClass,
        inputValueProvider.unconvertedValue);
  }

  private <S> S getLiteralInputValue(DataType<S> dataType, net.opengis.wps.v_1_0_0.DataType input) {
    LiteralType<S> literalType = context.getLiteralTypeRegistry().getLiteralType(dataType);
    return literalType.valueOf(input.getLiteralData().getValue());
  }

  private <S> S getComplexInputValue(DataType<S> dataType, net.opengis.wps.v_1_0_0.DataType input) {
    Format<S> format = context.getFormatRegistry().getDefaultFormat(dataType);
    return format.unmarshal(input.getComplexData().getContent());
  }
}
