/**
 * Legato is a configurable, lightweight web mapping client that can be
 * easily embedded into web pages and portals, CMS and individual web
 * applications. Legato is implemented in JavaScript and based on the
 * popular open source library OpenLayers.
 *
 * Copyright (C) 2010  disy Informationssysteme GmbH, http://www.disy.net
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.disy.ogc.wps.v_1_0_0.procedure;

import java.lang.reflect.Array;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Locale;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import net.disy.ogc.ows.v_1_1_0.OwsException;
import net.disy.ogc.wps.v_1_0_0.util.WpsUtils;
import net.opengis.wps.v_1_0_0.DataInputsType;
import net.opengis.wps.v_1_0_0.ExecuteResponse;
import net.opengis.wps.v_1_0_0.InputType;
import net.opengis.wps.v_1_0_0.ObjectFactory;
import net.opengis.wps.v_1_0_0.OutputDefinitionsType;
import net.opengis.wps.v_1_0_0.OutputDescriptionType;
import net.opengis.wps.v_1_0_0.ProcessDescriptionType;
import net.opengis.wps.v_1_0_0.ResponseFormType;
import net.opengis.wps.v_1_0_0.StatusType;
import net.opengis.wps.v_1_0_0.ExecuteResponse.ProcessOutputs;

public class ProcedureWpsProcessExecutor {

  private final WpsProcessContext context;

  public ProcedureWpsProcessExecutor(WpsProcessContext context) {
    this.context = context;
  }

  public ExecuteResponse execute(
      DataInputsType dataInputs,
      ResponseFormType responseForm,
      WpsProcedure processHelper) throws OwsException {
    ProcedureWpsProcessDescriptionFactory processDescriptionFactory = new ProcedureWpsProcessDescriptionFactory(
        context);
    ProcessDescriptionType processDescription = processDescriptionFactory
        .createProcessDescription(processHelper);

    final ExecuteResponse executeResponse = createResponse(
        processDescription,
        responseForm,
        dataInputs);

    Object[] arguments = getArguments(processHelper, dataInputs);
    Object result = processHelper.execute(arguments);
    executeResponse.setProcessOutputs(getOutputs(processHelper, processDescription, result));
    return executeResponse;
  }

  private ExecuteResponse createResponse(
      ProcessDescriptionType processDescription,
      ResponseFormType responseForm,
      DataInputsType dataInputs) {
    ObjectFactory wpsObjectFactory = context.getWpsObjectFactory();
    final ExecuteResponse executeResponse = wpsObjectFactory.createExecuteResponse();
    executeResponse.setService("WPS"); //$NON-NLS-1$
    executeResponse.setVersion("1.0.0"); //$NON-NLS-1$
    executeResponse.setProcess(WpsUtils.createBriefDescription(processDescription));

    final StatusType status = wpsObjectFactory.createStatusType();
    status.setProcessSucceeded("Ok."); //$NON-NLS-1$
    try {
      status.setCreationTime(DatatypeFactory.newInstance().newXMLGregorianCalendar(
          new GregorianCalendar()));
    }
    catch (DatatypeConfigurationException ignored) {
      // nothing to do
    }
    executeResponse.setStatus(status);

    // If lineage is true, data inputs should be returned
    final boolean lineage = responseForm != null
        && responseForm.getResponseDocument() != null
        && responseForm.getResponseDocument().isLineage();
    if (lineage) {
      executeResponse.setDataInputs(dataInputs);
      if (!responseForm.getResponseDocument().getOutput().isEmpty()) {
        final OutputDefinitionsType outputDefinitions = wpsObjectFactory
            .createOutputDefinitionsType();
        outputDefinitions.getOutput().addAll(responseForm.getResponseDocument().getOutput());
        executeResponse.setOutputDefinitions(outputDefinitions);
      }
    }

    executeResponse.setLang(Locale.GERMAN.toString());

    return executeResponse;
  }

  private Object[] getArguments(WpsProcedure processHelper, DataInputsType dataInputs) {
    List<InputParameterDescription> inputParameters = processHelper.getInputParameters();
    InputArgumentBuilder argumentBuilder = new InputArgumentBuilder(context);
    argumentBuilder.setInputParameters(inputParameters);

    if (dataInputs != null) {
      for (InputType input : dataInputs.getInput()) {
        argumentBuilder.addInput(input);
      }
    }

    return argumentBuilder.getArguments();
  }

  private ProcessOutputs getOutputs(
      WpsProcedure processHelper,
      ProcessDescriptionType processDescription,
      Object result) {
    ProcessOutputBuilder processOutputBuilder = new ProcessOutputBuilder(context);

    List<OutputDescriptionType> outputDescriptions = processDescription
        .getProcessOutputs()
        .getOutput();
    if (outputDescriptions.size() == 0) {
      return processOutputBuilder.getProcessOutputs();
    }

    OutputDescriptionType outputDescription = outputDescriptions.get(0);
    OutputParameterDescription outputParameter = processHelper.getOutputParameters().get(0);
    Class<?> outputClass = outputParameter.getParameterClass();
    if (result == null || !outputClass.isArray()) {
      processOutputBuilder.addOutput(outputDescription, outputParameter, outputClass, result);
    }
    else {
      for (int i = 0; i < Array.getLength(result); i++) {
        processOutputBuilder.addOutput(outputDescription, outputParameter, outputClass
            .getComponentType(), Array.get(result, i));
      }
    }

    return processOutputBuilder.getProcessOutputs();
  }
}
